;========
;=
;=  Loadable video driver for JCPIC.EXE
;=  Written by JC Moxon
;=  Copyright (c) 1996 Bitman Logic Corporation
;=
;=  Standard BIOS driver for just about any card.
;=  These functions are built in to JCPIC.EXE for VESA cards only.
;=
;========

.MODEL SMALL
.386

;=
;= Functions in each driver (with C declaration):
;=
;= int detect(void);
;= int open_drv(int mode);
;= void set_palette(int start, int count, char *palette);
;= void set_border(int color);
;= void set_pixel(int x, int y, int color);
;= void write_line(int x, int y, int width, unsigned char *line);
;= void scale_line(int x, int y, int awidth, int dwidth, unsigned char *line);
;= int close_drv(void);
;= void reserved_1(void);
;= void reserved_2(void);
;= void reserved_3(void);
;= void start_up(void);
;= void finish_up(void);
;=

.CODE
                org     0000h           ; Set location counter
dev_sig         db      'JCPIC_DRV'      ; Signature

;=
;= This routine is used to verify that the proper hardware is installed.
;= It also sets up the pointers to all of the other functions.
;=

_header     proc far

            dw   detect
            dw   ?
            dw   open_drv               ; Pointers to all of the functions.
            dw   ?                      ; These are filled in by JCPIC.EXE
            dw   set_palette            ; and shouldn't be moved.
            dw   ?
            dw   set_border
            dw   ?
            dw   set_pixel
            dw   ?
            dw   write_line
            dw   ?
            dw   scale_line
            dw   ?
            dw   close_drv
            dw   ?
            dw   reserved_1
            dw   ?
            dw   reserved_2
            dw   ?
            dw   reserved_3
            dw   ?
            dw   reserved_4
            dw   ?
            dw   start_up
            dw   ?
            dw   finish_up
            dw   ?

            screen_seg  dw 0A000h       ; Variables accessed by JCPIC.EXE
            card_mem    dw 256          ; Available memory on card.

            driver_version    dw 1
            driver_subversion dw 0
            reserved_int_1    dw ?
            reserved_int_2    dw ?
            reserved_int_3    dw ?
            reserved_int_4    dw ?
            driver_name       db 'BIOS Video Driver for JCPIC.EXE', 0
                              ; 32 byte device name (must be NULL terminated)



            ; Driver Info
            modes             dw 0013h, 0061h,  0053h,  0054h,  0059h,  005Eh
            enabled           db 0,    0,  0,  0,  0,  0
            max_x             dw 320,  600,  640,  800,  1024,  1280
            max_y             dw 200,  400,  480,  600,  768,  1024

            name_0            db 'User Defined Mode', 0    ;64k
            name_1            db '640 x 400 x 256  ', 0    ;256k
            name_2            db '640 x 480 x 256  ', 0    ;512k
            name_3            db '800 x 600 x 256  ', 0    ;512k
            name_4            db '1024 x 763 x 256 ', 0    ;1024k
            name_5            db '1280 x 1024 x 256', 0    ;2048k

            ; Ok to add globals here
            curr_mode         db ?
            ax_val            dw 0
            bx_val            dw 0
	    input             db 10 dup(?)
	    mode_prompt       db 13, 10, 'Enter AX value (mode number, 4 hex digits):  ', '$'
	    fcn_prompt        db 13, 10, 'Enter BX value (0 is default):  ', '$'
	    x_prompt          db 13, 10, 'Enter X dimension:  ', '$'
	    y_prompt          db 13, 10, 'Enter Y dimension:  ', '$'
            valid_keys        db '0123456789ABCDEFabcdef'
            x                 db 0
            y                 db 0
            pos               db 0

            intro_msg         db 13, 10, 'Standard BIOS JCPIC display driver written by JC Moxon'
                              db 13, 10, 'Copyright (c) 1996 Bitman Logic Corporation', 13, 10
                              db 13, 13, 10, 'Press any key to continue...', '$'
            retf
            endp

;=
;= Macro to write a '$' terminated string to screen
;=

write_string    MACRO   __msg

                push ax
                push dx
                push ds

                push cs
                pop  ds

                mov ax, 0900h
                mov dx, offset __msg
                int 021h

                pop ds
                pop dx
                pop ax

ENDM

;=
;= Procedure to convert from a hex string to a hex digit
;= Converts until it hits 0Dh, only takes last 4 digits
;=

hexstr_to_ax    proc  near                             ;Each digit is 4 bits

                mov    bx, 0
                mov    ax, 0
   hcon_loop:
                cmp    input[bx], 0Dh
                je     hcon_don

                cmp    input[bx], 'a'
                jb     his_upper

                sub    input[bx], 'a'                      ;Convert to alpha
                add    input[bx], '0'
                add    input[bx], 10
                jmp    his_digit

   his_upper:
                cmp    input[bx], 'A'
                jb     his_digit

                sub    input[bx], 'A'
                add    input[bx], '0'
                add    input[bx], 10

   his_digit:
                mov    ch, input[bx]
                sub    ch, '0'
                shl    ax, 4

                or     al, ch

                inc    bx
                cmp    bx, 10
                ja     hcon_don

                jmp    hcon_loop

     hcon_don:
                push   ax

                mov    cx, 5
                mov    ax, 0
                mov    di, offset input
                rep    stosw

                pop    ax

                ret
                endp

;=
;= Procedure to retrieve up to 10 hex digits from keyboard
;=

hex_string     proc near

                mov    ah, 03h                          ; Store cursor position
                mov    bh, 0
                int    010h
                mov    x,  dl
                mov    y,  dh

                xor    bx, bx
                cld                                     ; Set direction flag
                mov    pos, 0
                push   cs
                pop    es

  hkey_loop:
                mov  ah, 1                              ; Read AX one hex
                int  021h                               ; digit at a time

                cmp  al, 008h                           ; Look for backspace
                je   hbackspace

                cmp   al, 0dh                           ; Done
                je    hgs_done

                mov  bl, pos
                mov  input[bx], al

                inc  pos                                ; Keep in bounds
                inc  x

                mov  cx, 23
                mov  di, offset valid_keys              ; Make sure key is
                repnz scasb                             ; valid hex digit

                jcxz  hbackspace

                cmp  pos, 10
                jb   hkey_loop

   hbackspace:
                cmp  pos, 0                             ; Keep from backspacing
                je   hleft                              ; over prompt

                dec  pos
                dec  x

        hleft:
                mov  ah, 02h                            ; Reposition cursor
                mov  bh, 0
                mov  dh, y
                mov  dl, x
                int  010h

                mov  dl, ' '                            ; Erase character
                mov  ah, 2
                int  021h

                mov  ah, 02h                            ; Reposition cursor
                mov  bh, 0
                mov  dh, y
                mov  dl, x
                int  010h

                jmp  hkey_loop

       hgs_done:

                mov  bl, pos                            ; Mark end of input
                mov  input[bx], 0Dh                     ; with 0D

                ret
                endp

;=
;= Procedure to retrieve up to 10 digits from keyboard
;=

dec_string     proc near

                mov    ah, 03h                          ; Store cursor position
                mov    bh, 0
                int    010h
                mov    x,  dl
                mov    y,  dh

                xor    bx, bx
                cld                                     ; Set direction flag
                mov    pos, 0
                push   cs
                pop    es

  key_loop:
                mov  ah, 1                              ; Read AX one hex
                int  021h                               ; digit at a time

                cmp  al, 008h                           ; Look for backspace
                je   backspace

                cmp   al, 0dh                           ; Done
                je    gs_done

                mov  bl, pos
                mov  input[bx], al

                inc  pos                                ; Keep in bounds
                inc  x

                mov  cx, 11
                mov  di, offset valid_keys              ; Make sure key is
                repnz scasb                             ; valid digit

                jcxz  backspace

                cmp  pos, 10
                jb   key_loop

   backspace:
                cmp  pos, 0                             ; Keep from backspacing
                je   left                               ; over prompt

                dec  pos
                dec  x

        left:
                mov  ah, 02h                            ; Reposition cursor
                mov  bh, 0
                mov  dh, y
                mov  dl, x
                int  010h

                mov  dl, ' '                            ; Erase character
                mov  ah, 2
                int  021h

                mov  ah, 02h                            ; Reposition cursor
                mov  bh, 0
                mov  dh, y
                mov  dl, x
                int  010h

                jmp  key_loop

       gs_done:

                mov  bl, pos                            ; Mark end of input
                mov  input[bx], 0Dh                     ; with 0D

                ret
                endp

;=
;= Procedure to convert from a decimal string to a dec digit
;= Converts until it hits 0Dh, only takes last 4 digits
;=

decstr_to_ax    proc near                                  ;Each digit is 4 bits

                mov    bx, 0
                mov    ax, 0
   con_loop:
                cmp    input[bx], 0Dh
                je     con_don

                mov    ch, 10
                mul    ch

                mov    ch, input[bx]
                sub    ch, '0'

                add    al, ch

                inc    bx
                cmp    bx, 10
                ja     con_don

                jmp    con_loop

     con_don:
                push   ax

                mov    cx, 5
                mov    ax, 0
                mov    di, offset input
                rep    stosw

                pop    ax

                ret
                endp

;=
;= Routine to verify that hardware is present.
;= Should return 0 in AX if present, -1 if not.
;=

detect          proc near

                mov    ax, 0100h
                mov    cx, 0D0Eh                        ; Turn cursor on
                int    010h

                write_string mode_prompt
                call   hex_string
                call   hexstr_to_ax
                mov    ax_val, ax

                write_string fcn_prompt
                call   hex_string
                call   hexstr_to_ax
                mov    bx_val, ax

                write_string x_prompt
                call   dec_string
                call   decstr_to_ax
                mov    bx, 0
                mov    max_x[bx], ax

                write_string y_prompt
                call   dec_string
                call   decstr_to_ax
                mov    bx, 0
                mov    max_y[bx], ax

                mov    ax, 0100h                        ; Turn cursor off
                mov    cx, 2000h
                int    010h

		mov enabled[0], 1
		mov enabled[1], 0
		mov enabled[2], 0
		mov enabled[3], 0
		mov enabled[4], 0
		mov enabled[5], 0

                mov ax, 0
                retf

                endp

;=
;= Function to set mode and return result.
;=

open_drv        proc near

                mov  ax, ax_val
                mov  bx, bx_val
		int  010h

		mov  ax, 0
		retf
                endp

;=
;= Routine to set VGA palette registers
;=

set_palette     proc near

		push    bp
		mov     bp,sp

                mov ax, 01012h
                mov bx, [bp + 6]
                mov cx, [bp + 8]
                les dx, [bp + 10]

                int 010h

		pop     bp
		retf
		endp

;=
;= Procedure to set screen border and background
;=

set_border      proc near

		push bp
		mov  bp, sp

		mov ax, 01001h         ; Prepare to set border using ISR
		mov bx, [bp + 6]       ; Store color to BL
		xor bh, bh
		int 010h               ; Set it

		pop  bp
		retf
		endp

;=
;= Procedure to set individual pixel
;=

set_pixel       proc near

		push bp
		mov  bp, sp

                mov ax, [bp + 10]
                mov ah, 0Ch
                mov cx, [bp + 6]
                mov dx, [bp + 8]
                int 010h

		pop  bp
		retf
		endp

;=
;= Procedure to write one line of memory to screen
;=

write_line      proc near

		push bp
		mov  bp, sp

                push si
                push di
                push ds

                cld
                mov        cx, [bp + 6]     ; X value
                mov        dx, [bp + 8]     ; Y value
                mov        di, [bp + 10]    ; Width to DI
                lds        si, [bp + 12]    ; Load pointer
      bl_loop:
                mov        ah, 0Ch          ; Function 0Ch
                lodsb                       ; Mov al, ds:[si]
                int        010h             ; Set pixel
                inc        cx               ; Next x position
                dec        di
                jnz        bl_loop

                pop  ds
                pop  di
                pop  si

                pop  bp
		retf
		endp

;=
;= Procedure to scale one line of memory onto screen.
;= Uses fixed point math in 32 bit registers with high 24 bits holding
;= offset and low 8 bits holding decimal place.
;=

scale_line      proc near

  X_VAL   EQU [bp + 6]
  Y_VAL   EQU [bp + 8]
  A_WIDTH EQU [bp + 10]
  D_WIDTH EQU [bp + 12]
  PTR_OFF EQU [bp + 14]
  PTR_SEG EQU [bp + 16]
  X_STEP  EQU dword ptr [bp + 18]

                push bp
                mov  bp, sp

                push si
                push di
                push ds

                mov        dx, 0
                mov        ax, A_WIDTH      ; Calculate the width of each pixel
                mov        bx, D_WIDTH
                div        bx
                shl        eax, 16
                mov        ax, dx
                div        bx
                shr        eax, 8
                mov        X_STEP, eax

                mov        cx, X_VAL        ; X value
                mov        di, D_WIDTH      ; Width to DI
                mov        esi, 0
                lds        si, PTR_OFF      ; Load our points
                mov        edx, 0
                mov        ebx, 0
   bs_loop:
                mov        ah, 0Ch          ; Function to AH
                mov        al, ds:[si + bx] ; Color to AL
                add        edx, X_STEP      ; Fixed point math
                mov        ebx, edx
                shr        ebx, 8           ; Next pixel in DS:[SI + BX]
                push       edx              ; Save our offset
                mov        dx, Y_VAL        ; Y value to DX
                int        010h
                pop        edx

                inc        cx               ; Next pixel
                dec        di
                jnz        bs_loop

                pop  ds
                pop  di
                pop  si
                pop  bp
		retf
		endp



;=
;= Procedure to close down and return to text mode
;=

close_drv       proc near

                mov  ax, 00003h
                int  010h
                mov  ax, 0

                retf
                endp

;=
;= These functions are reserved space.
;=
reserved_1  proc near
            retf
            endp
reserved_2  proc near
            retf
            endp
reserved_3  proc near
            retf
            endp
reserved_4  proc near
            retf
            endp

;=
;= Routine to do anything you like with!  Great for advertisements!!!
;= Called immediately after driver is loaded at beginning of program.
;=

start_up        proc near

                retf
                endp

;=
;= Routine to do anything you like with!  Great for advertisements!!!
;= Called just before driver is unloaded at end of program.
;=

finish_up       proc near

                retf
                endp

End
;EOF -JCM- ;
