;========
;=
;=  Loadable video driver for JCPIC.EXE
;=  Written by JC Moxon
;=  Copyright (c) 1996 Bitman Logic Corporation
;=
;=  DRV.ASM - Debugging driver.
;=  These functions are built in to JCPIC.EXE for VESA cards only.
;=
;========

.MODEL SMALL
.386

;=
;= Functions in each driver (with C declaration):
;=
;= int detect(void);
;= int open_drv(int mode);
;= void set_palette(int start, int count, char *palette);
;= void set_border(int color);
;= void set_pixel(int x, int y, int color);
;= void write_line(int x, int y, int width, unsigned char *line);
;= void scale_line(int x, int y, int awidth, int dwidth, unsigned char *line);
;= int close_drv(void);
;= void reserved_1(void);
;= void reserved_2(void);
;= void reserved_3(void);
;= void start_up(void);
;= void finish_up(void);
;=

.CODE
                org     0000h           ; Set location counter
dev_sig         db      'JCPIC_DRV'      ; Signature

;=
;= This routine is used to verify that the proper hardware is installed.
;= It also sets up the pointers to all of the other functions.
;=

_header     proc far

            dw   detect
            dw   ?
            dw   open_drv               ; Pointers to all of the functions.
            dw   ?                      ; These are filled in by JCPIC.EXE
            dw   set_palette            ; and shouldn't be moved.
            dw   ?
            dw   set_border
            dw   ?
            dw   set_pixel
            dw   ?
            dw   write_line
            dw   ?
            dw   scale_line
            dw   ?
            dw   close_drv
            dw   ?
            dw   reserved_1
            dw   ?
            dw   reserved_2
            dw   ?
            dw   reserved_3
            dw   ?
            dw   reserved_4
            dw   ?
            dw   start_up
            dw   ?
            dw   finish_up
            dw   ?

            screen_seg  dw 0A000h       ; Variables accessed by JCPIC.EXE
            card_mem    dw 256          ; Available memory on card.

            driver_version    dw 1
            driver_subversion dw 0
            reserved_int_1    dw ?
            reserved_int_2    dw ?
            reserved_int_3    dw ?
            reserved_int_4    dw ?
            driver_name       db 'Paradise/Western Digital Driver', 0
                              ; 32 byte device name (must be NULL terminated)


            ; Driver Info
            modes             dw 0013h, 005Eh,  005Fh,  005Ch,  0060h,  0000h
            enabled           db 0,    0,  0,  0,  0,  0
            max_x             dw 320,  640,  640,  800,  1024,  1280
            max_y             dw 200,  400,  480,  600,  768,  1024

            name_0            db '320 x 200 x 256  ', 0    ;64k
            name_1            db '640 x 400 x 256  ', 0    ;256k
            name_2            db '640 x 480 x 256  ', 0    ;512k
            name_3            db '800 x 600 x 256  ', 0    ;512k
            name_4            db '1024 x 763 x 256 ', 0    ;1024k
            name_5            db '1280 x 1024 x 256', 0    ;2048k

            ; Ok to add globals here
            curr_mode         db ?
            Current_Bank      db 0
            old_reg           db ?
            paradise_msg      db 'Paradise or Western Digital VGA card ', '$'
            has_2_msg         db 'with 2 MB of memory.', '$'
            has_1_msg         db 'with 1 MB of memory.', '$'
            has_512_msg       db 'with 512 kb of memory.', '$'
            has_256_msg       db 'with 256 kb of memory.', '$'
            intro_msg         db 13, 10, 'Paradise/Western Digital VGA JCPIC display driver written by JC Moxon'
                              db 13, 10, 'Copyright (c) 1996 Bitman Logic Corporation'
                              db 13, 10, 10, 'Special thanks to Finn Thoegersen for the VGADOC publication.'
                              db 13, 10, 'This driver wouldn''t have been possible without it.  The publication is'
                              db 13, 10, 'available on the Game Developers forum of Compuserve.'
                              db 13, 10, 10, 'Press any key to continue...', '$'
            retf
            endp

TestPort	proc	near
		push	cx

		in	al,dx
		mov	cl,al		;; old
		mov	bh,bl
		not	bh
		and	al,bh
		out	dx,al
		in	al,dx
		and	al,bl
		jnz	@@NoPort
		mov	al,cl
		or	al,bl
		out	dx,al
		in	al,dx
		and	al,bl
		cmp	al,bl
@@NoPort:	mov	al,cl
		out	dx,al
		pop	cx
		ret
TestPort	endp

TestPortIndexed	proc near
		out	dx,al
		inc	dx
		call	TestPort
		ret
TestPortIndexed	endp

rdinx		MACRO	__port,__inx
		mov	dx,__port
		mov	al,__inx
		out	dx,al
		inc	dx
		in	al,dx
ENDM

tstrg		MACRO	__port,__mask
		push	bx
		mov	bl,__mask
		mov	dx,__port
		call	TestPort
		pop	bx
ENDM

testinx2	MACRO	__port,__reg,__mask
		push	bx
		mov	bl,__mask
		mov	al,__reg
		mov	dx,__port
		call	TestPortIndexed
		pop	bx
ENDM

testinx 	MACRO	__port,__reg
		push	bx
		mov	bl, 0FFh
		mov	al, __reg
		mov	dx, __port
		call	TestPortIndexed
		pop	bx
ENDM

wrinx		MACRO	__port,__inx,__val
		push	ax
		mov	dx,__port
		mov	al,__inx
		out	dx,al
		inc	dx
		pop	ax
		mov	al,__val
		out	dx,al
ENDM

tstreg		MACRO	__port,__mask
		push	bx
		mov	bl,__mask
		mov	dx,__port
		call	TestPort
		pop	bx
ENDM

setinx          MACRO   __port,__inx,__val
                push    ax
                rdinx   __port, __inx
                or      al, __val
                wrinx   __port, __inx, al
                pop     ax
ENDM

modinx		MACRO	__port,__inx,__mask,__nwv
		push	ax
		mov	dx,__port
		mov	ah,__nwv
		and	ah,__mask
		mov	al,__inx
		out	dx,al
		inc	dx
		in	al,dx
		and	al,not __mask
		or	al,ah
		out	dx,al
		pop	ax
ENDM

write_string    MACRO   __msg

                push ax
                push dx
                push ds

                push cs
                pop  ds

                mov ax, 0900h
                mov dx, offset __msg
                int 021h

                pop ds
                pop dx
                pop ax

ENDM

paradise_bank   proc near
		push	ax
		push	dx
		mov	ax,dx
		mov	[Current_Bank],al
		cli

	        push	cx
		mov	cl,4
		shl	al,cl
		wrinx	03ceh,9,al
		wrinx	03ceh,0ah,al
		pop	cx

		sti
		pop	dx
		pop	ax
		ret

                endp

;=
;= Routine to verify that hardware is present.
;= Should return 0 in AX if present, -1 if not.
;=

detect          proc near

                rdinx	03ceh, 0fh
		mov	old_reg, al
		setinx	03ceh, 0fh, 017h, 0	;; Lock registers
		testinx	03ceh, 9, 07fh
		jz	not_paradise
		modinx	03ceh, 0fh, 17h, 05     ;; Unlock registers
		testinx	03ceh, 9, 07fh
		jnz	is_paradise

 not_paradise:
                wrinx	03ceh, 0fh, old_reg
                mov     ax, -1
                retf

  is_paradise:
                write_string paradise_msg

                rdinx  03CEh, 0Bh               ;; Check for memory
                shr    al, 6
                and    al, 3

                cmp    al, 2
                je     has_512
                cmp    al, 3
                je     has_1MB

  has_256:
                mov enabled[0], 1     ;320 x 200
                mov enabled[1], 1     ;640 x 400
                mov enabled[2], 0     ;640 x 480
                mov enabled[3], 0     ;800 x 600
                mov enabled[4], 0     ;1024 x 768
                mov enabled[5], 0     ;1280 x 1024

                write_string has_256_msg

                mov ax, 0
                retf

  has_2MB:
                mov enabled[0], 1     ;320 x 200
                mov enabled[1], 1     ;640 x 400
                mov enabled[2], 1     ;640 x 480
                mov enabled[3], 1     ;800 x 600
                mov enabled[4], 1     ;1024 x 768
                mov enabled[5], 0     ;1280 x 1024

                write_string has_2_msg

                mov card_mem,  2048
                mov ax, 0
                retf
  has_1MB:
                mov enabled[0], 1     ;320 x 200
                mov enabled[1], 1     ;640 x 400
                mov enabled[2], 1     ;640 x 480
                mov enabled[3], 1     ;800 x 600
                mov enabled[4], 1     ;1024 x 768
                mov enabled[5], 0     ;1280 x 1024

                write_string has_1_msg

                mov card_mem,  1024
                mov ax, 0
                retf

  has_512:
                mov enabled[0], 1     ;320 x 200
                mov enabled[1], 1     ;640 x 400
                mov enabled[2], 1     ;640 x 480
                mov enabled[3], 1     ;800 x 600
                mov enabled[4], 0     ;1024 x 768
                mov enabled[5], 0     ;1280 x 1024

                write_string has_512_msg

                mov card_mem,  512
                mov ax, 0
                retf

  not_oak:
                mov ax, -1
                retf
                endp

;=
;= Function to set mode and return result.
;=

open_drv        proc near

                push bp                     ; Address stack through BP
                mov  bp, sp

                mov  bx, [bp + 6]           ; Mode in [BP + 6]
                shl  bx, 1
                mov  ax, modes[bx]
                mov  curr_mode, bl
                cmp  ax, 0                  ; Something is missing
                jne  is_good
                mov  ax, -1
                retf

  is_good:
                int  010h

                mov  ax, 0
                pop  bp
                retf
                endp

;=
;= Routine to set VGA palette registers
;=

set_palette     proc near
	        push	bp
		mov	bp,sp

		mov     dx,03DAH       ;Wait for end of
Vscan:                                 ;vertical rescan
		in      al,dx
		test    al,8
		jne     Vscan

Rscan:
		in      al,dx          ;Go to start of rescan
		test    al,8
		je      Rscan


		mov	dx,03C6H			;Set Pel mask to 0xFF
		mov	al,0FFH
		out	dx,al

		mov	dx,03C8H			;Load Write address
		mov	al,byte ptr [bp+6]		;with starting number
		out	dx,al

		xor	cx,cx
		mov	dx,03C9H			;Load dx with register
		les	bx,dword ptr [bp+10]            ;Load al with Red
SetNext:
		mov	al,byte ptr es:[bx]

		out	dx,al				;Set Red

		mov	al,byte ptr es:[bx+1]           ;Set Green
		out	dx,al

		mov	al,byte ptr es:[bx+2]           ;Set Blue
		out	dx,al

		add	word ptr bx,3              ;Move to next DAC
		inc	cx

		cmp	cx,word ptr [bp+8]		;See if we changed
		jl	SetNext				;enough of them

		pop	bp
                retf
                endp

;=
;= Procedure to set screen border and background
;=
set_border      proc near

                push bp
                mov  bp, sp

	        mov	es, screen_seg

                mov        [Current_Bank], 0  ;Reset to first bank
                call       paradise_bank

                mov        dx, 0
                mov        ax, 65535        ; Lines per bank
                mov        bl, curr_mode
                xor        bh, bh
                div        max_x[bx]
                push       dx

                mov        dx, 0
                push       ax               ; Banks per screen
                mov        ax, max_x[bx]
                pop        bx
                div        bx

                mov        bx, ax
                pop        dx
                cmp        dx, 0
                je         no_mod
                inc        bx

      no_mod:
                mov        di, 0
      loop_y:
                dec        bx
                mov        dx, bx
                cmp        dl, Current_Bank
                je         skip_bank
                call       paradise_bank
     skip_bank:
                mov        ax, [bp + 6]    ; Color
                mov        ah, al          ; Duplicate color 4 times
                mov        cx, ax
                shl        eax, 16
                mov        ax, cx
                mov        cx, 16384       ; Clear 64k of video RAM
	        rep        stosd           ; Store Pixels 4 at a time

                cmp        bx, 0
                ja         loop_y

      complete:

                mov ax, 01001h
                mov bx, [bp + 6]
                xor bh, bh
                int 010h

                pop  bp
                retf

                endp

;=
;= Procedure to set individual pixel
;=
set_pixel       proc near

                push bp
                mov  bp, sp

                mov     cx, [bp + 6]     ; X value
                mov     ax, [bp + 8]     ; Y value
                mov     bl, curr_mode
                xor     bh, bh
	        mul	max_x[bx]	 ; Calculate offset of pixel
	        add	cx, ax
                adc     dl, 0
	        cmp	dl, Current_Bank ; Check if pixel is in current bank
	        je	snonew
	        call	paradise_bank
   snonew:
	        mov	es, screen_seg
                mov     dx, [bp + 10]    ; Color
	        mov	es:[bx], dl	 ; Store Pixel

                pop  bp
                retf

;=
;= Procedure to write one line of memory to screen
;=
write_line      proc near

                push bp
                mov  bp, sp

                push si
                push di
                push ds

                cld                      ; Set direction flag
                mov     di, [bp + 6]     ; X value
                mov     ax, [bp + 8]     ; Y value
                mov     bl, curr_mode
                xor     bh, bh
	        mul	max_x[bx]	 ; Calculate offset of line
	        add	di, ax
                adc     dl, 0
	        cmp	dl, Current_Bank ; Check if offset is in current bank
	        je	lnonew
	        call	paradise_bank
   lnonew:
	        mov	es, screen_seg
                mov     cx, [bp + 10]    ; Width to CX
                shr     cx, 2
                lds     si, [bp + 12]
                mov     ax, 0
                sub     ax, [bp + 10]
                cmp     di, ax
                ja      broken
                rep     movsd

                pop  ds
                pop  di
                pop  si
                pop  bp
                retf

   broken:
                shl     cx, 2           ; Round off to nearest 4 byte boundary
   b_loop:
                cmp     di, 0
                ja      no_bank
                pop     ds
                mov     dl, [Current_Bank]
                inc     dl
                call    paradise_bank
                push    ds
                mov     ds, [bp + 14]
   no_bank:
                movsb
                loop    b_loop

                pop  ds
                pop  di
                pop  si
                pop  bp
                retf
                endp

;=
;= Procedure to scale one line of memory onto screen.
;= Uses fixed point math in 32 bit registers with high 24 bits holding
;= offset and low 8 bits holding decimal place.
;=
scale_line      proc near

  X_VAL   EQU [bp + 6]
  Y_VAL   EQU [bp + 8]
  A_WIDTH EQU [bp + 10]
  D_WIDTH EQU [bp + 12]
  PTR_OFF EQU [bp + 14]
  PTR_SEG EQU [bp + 16]
  X_STEP  EQU dword ptr [bp + 18]

             push bp
             mov  bp, sp

             push si
             push di
             push ds

             cld                         ; Set direction flag
             mov        dx, 0
             mov        ax, A_WIDTH      ; Calculate the width of each pixel
             mov        bx, D_WIDTH
             div        bx
             shl        eax, 16
             mov        ax, dx
             div        bx
             shr        eax, 8
             mov        X_STEP, eax

             mov        di, X_VAL        ; X value
             mov        ax, Y_VAL        ; Y value
             mov        bl, curr_mode
             xor        bh, bh
	     mul	max_x[bx]	 ; Calculate offset of line
	     add	di, ax
             adc        dl, 0
	     cmp	dl, Current_Bank ; Check if offset is in current bank
	     je	        spnonew
	     call	paradise_bank
spnonew:
	     mov	es, screen_seg
             mov        cx, D_WIDTH      ; Width to CX
             lds        si, PTR_OFF
             mov        ax, 0
             sub        ax, D_WIDTH
             mov        edx, 0
             mov        ebx, 0
             cmp        di, ax
             ja         broken_loop
sloop:
             mov        al, byte ptr [si + bx]
             stosb
             add        edx, X_STEP
             mov        ebx, edx
             shr        ebx, 8
             loop       sloop

             pop  ds
             pop  di
             pop  si
             pop  bp
             retf
             endp

broken_loop:
             cmp        di, 0
             ja         sno_bank
             pop        ds
             mov        dl, [Current_Bank]
             inc        dl
             call       paradise_bank
             push       ds
             mov        ds, PTR_SEG
sno_bank:
             mov        al, byte ptr [si + bx]
             stosb
             add        edx, X_STEP
             mov        ebx, edx
             shr        ebx, 8
             loop       broken_loop

             pop  ds
             pop  di
             pop  si
             pop  bp
             retf
             endp

;=
;= Procedure to close down and return to text mode
;=
close_drv       proc near

                mov  ax, 00003h
                int  010h
                mov  ax, 0

                retf
                endp

;=
;= These functions are reserved space.
;=
reserved_1  proc near
            retf
            endp
reserved_2  proc near
            retf
            endp
reserved_3  proc near
            retf
            endp
reserved_4  proc near
            retf
            endp

;=
;= Routine to do anything you like with!  Great for advertisements!!!
;= Called immediately after driver is loaded at beginning of program.
;=
start_up        proc near

                mov ax, 0003h
                int 010h

                write_string intro_msg

                mov ax, 0001h
                int 016h

                retf
                endp

;=
;= Routine to do anything you like with!  Great for advertisements!!!
;= Called just before driver is unloaded at end of program.
;=
finish_up       proc near

                retf
                endp

End
;EOF -JCM- ;
